#include <Adafruit_Arcada.h>
#include <Adafruit_NeoPixel.h>
#include "car_sport.h"
#include "car_cabrio.h"
#include "car_f1.h"
#include "background_garage.h"
#include "background_city.h"
#include "background_fields.h"
#include "background_beach.h"


//////////////////////////////////////////////// Constants define ////////////////////////////////////////////////

#define HIGHSCORESPATH "/highscores"
#define USERSETTINGSPATH "/user-settings.json"
#define JSONEXTENSION ".json"

#define MAINMENUSIZE 5
#define PLAYMENUSIZE 3
#define SETTINGSMENUSIZE 4


//////////////////////////////////////////////// Global variables ////////////////////////////////////////////////

// Common

Adafruit_Arcada arcada;

int programState = 0; // State of the gameflow. 0 = menu, 1 = game, 2 = scoreboard

uint8_t buttonsPressed = 0;
uint8_t buttonsHold = 0;

Adafruit_NeoPixel pixels(5, 8, NEO_GRB + NEO_KHZ800);


// User settings

StaticJsonDocument<256> configJSON;

String brightnessDisplayLabels[] = {"Full", "Low", "Lowest"};
uint8_t brightnessDisplayValues[] = {255, 100, 20};
int brightnessDisplayIndex = 0;

String brightnessLEDLabels[] = {"BURN MY EYES", "High", "Medium", "Low"};
uint8_t brightnessLEDValues[] = {255, 60, 10, 1};
int brightnessLEDIndex = 2;

String volumeLabels[] = {"On", "Off"};
bool volumeOn = true;


// Menu

int menuIndexFirstLevel = 0;
int menuIndexSecondLevel = -1;
int menuIndexThirdLevel = -1;

String mainMenu[] = {"Play", "Garage", "Settings", "Scoreboard", "Credits"};

String playMenu[] = {"City", "Fields", "Beach"};

String settingsMenu[] = {"Brightness (Display)", "Brightness (LED)", "Volume", "Reset Scoreboard"};


// Game

double distanceDriven = 0;

bool temporarilyEnableThisCode = true; //temporarily activates the lower part of the code since the upper one wasn't carried out

int rev; // Area of rev counter

double x; // X-axis time passed

// Coefficient of speed function
double f1 = 0.03344120109;
double f2 = 1.3505210882678;
double f3 = 24.6757998921767;
double f4 = 15.7444624385786;

double y; // Counts mili-seconds in limiter, gets subtracted from x to limit the speed function

bool lastShiftWasLimited;

int gear = 0;

double distanceIndicatorProgress = 0;

int maxDistance = 1200;

uint16_t score;

double carSpeed;

int helpCalculateDrawBackground;

int helpCalculateY;

int helpCalculateReactionTime;

int helpCalculateDistanceDriven;

int helpCalculateStartOfGameplayTime;

bool resetHelper; //resets helpers that need to be reset at the start of the game (helpCalculateStartOfGameplayTime)

uint8_t carGearsAmount[3] = {7, 6, 8};

double maxSpeedPerGear[3][8] = {
    {85, 125, 172, 205, 244, 275, 320, -1},  //Sportscar
    {70, 120, 165, 195, 227, 250, -1, -1},   // Cabrio
    {108, 132, 158, 186, 228, 258, 295, 370} //F1 Car
};

int tooEarlyVariable[3][7] = {
    {735, 500, 363, 304, 256, 227, -1}, //Sportscar
    {892, 520, 378, 320, 275, -1, -1},  // Cabrio
    {578, 473, 395, 336, 274, 242, 211} //F1 Car
};

double multiplyxwith[] = {   //Multiplies the standard acceleration with the corresponding factor
    1.2,                     //Sportscar
    1,                       //Cabrio
    1.3,                     //F1 car
};

int measuredReactionTime[7][1];

int averageReactionTime;

int16_t xPosBackground = 0;
int16_t yPosBackground = 0;
int16_t xPosCar = 10;
int16_t yPosCar = 105;

String carNames[] = {"Sports Car", "Cabriolet", "F1 Car"};
const uint16_t *carGraphics[3] = {car_sport, car_cabrio, car_f1};
const uint8_t *carMasks[3] = {car_sport_mask, car_cabrio_mask, car_f1_mask};
int8_t selectedCarIndex = 0;

int8_t selectedBackground = 0;

bool startUpGame = false;
unsigned long timePassed = 0;


// Scoreboard

StaticJsonDocument<256> highscoreJSON;
uint16_t highscores[5] = {};

int8_t newHighscorePosition = -1;
uint8_t selectedScoreboardCarIndex = 0;


// Audio

File musicFile;
bool musicIsPlaying = false;
uint32_t bpmTiming = 400;
uint32_t beatTime;


// Credits

String creditsTexts[] = {
    "This game has been",
    "created during the 2022",
    "Multimodal Media Madness",
    "(M3) practical lab at",
    "RWTH Aachen by:",
    "",
    "Ilinca Baicu",
    "Jan Burda",
    "Lars Dirschauer",
    "",
    "Libraries used:",
    "Adafruit Arcada",
    "",
    "Image/Audio Credits to:",
    "clipart-library.com",
    "opengameart.org",
    "downloads.khinsider.com",
    //"1234567891234567891234567",
};
uint8_t creditsIndex = 0;


//////////////////////////////////////////////// Setup ////////////////////////////////////////////////

void setup() {

    // Initialize serial connection
    Serial.begin(9600);

    // Load Arcada library
    if (!arcada.arcadaBegin()) {
        Serial.println("Error: Failed to load Arcada library.");
    }

    // Initialize filesystem
    if (!arcada.filesysBegin()) {
        Serial.println("Error: Failed to load File system.");
    }

    // Load user settings
    loadUserSettings();

    // Setup display
    arcada.displayBegin();
    arcada.setBacklight(brightnessDisplayValues[brightnessDisplayIndex]);
    arcada.createFrameBuffer(ARCADA_TFT_WIDTH, ARCADA_TFT_HEIGHT);

    // Setup neopixels
    pixels.begin();
    pixels.setBrightness(brightnessLEDValues[brightnessLEDIndex]);

    // Setup game
    changeProgramState(0);

    Serial.println("Finished booting!");
}


//////////////////////////////////////////////// Loop ////////////////////////////////////////////////

void loop() {
    // Read buttons
    buttonsPressed = arcada.readButtons();

    // Enable music
    if (!musicIsPlaying) {
        loadMusic();
    }

    // fill music buffer if needed
    if (musicIsPlaying) {
        if (arcada.WavReadyForData()) {
            wavStatus status = arcada.WavReadFile();
        }
    }

    // Program flow
    switch (programState) {
        case 0:
            handleMenu();
            break;
        case 1:
            handleGameplayWrapper();
            break;
        case 2:
            handleScoreboard();
            break;
        default:
            Serial.println("Error: Variable 'programState' out of range: " + String(programState));
            break;
    }

    // update display
    arcada.blitFrameBuffer(0, 0, true);

    // perform standard board functions
    yield();
}


//////////////////////////////////////////////// Common helpers ////////////////////////////////////////////////

/**
 * @brief Calculates the index of a button mask, defined by the position of the single 1 in an int.
 * @param buttonMask An int with exactly one bit on 1.
 * @return An unsigned int with the position of the 1 bit.
 */
uint8_t getButtonMaskIndex(uint8_t buttonMask) {
    uint8_t index = 0;
    while (buttonMask != 0) {
        buttonMask = buttonMask >> 1;
        index++;
    }
    return (index - 1);
}

/**
 * @brief Returns true, if the button of the given mask has just been pressed.
 * @param buttonMask Similar format to ARCADA_BUTTONMASK_*. May only have a single 1 in it.
 */
bool buttonPressed(uint8_t buttonMask) {
    if (buttonsPressed & buttonMask) {
        if (!(buttonsHold & buttonMask)) {
            buttonsHold |= (1 << getButtonMaskIndex(buttonMask));
            return true;
        }
    } else {
        buttonsHold &= ~(1 << getButtonMaskIndex(buttonMask));
    }
    return false;
}

/**
 * @brief This function changes the state of the program to another. It resets all necessary variables and executes the change.
 * @param state Index of program state to change to.
 */
void changeProgramState(int state) {
    switch (state) {
        case 0:
            Serial.println("Info: Preparing for program state change to 'main menu'.");
            menuIndexFirstLevel = 0;
            menuIndexSecondLevel = -1;
            menuIndexThirdLevel = -1;
            newHighscorePosition = -1;
            programState = 0;
            Serial.println("Info: Change to program state 'main menu' completed.");
            break;
        case 1:
            Serial.println("Info: Preparing for program state change to 'gameplay'.");
            distanceDriven = 0;
            temporarilyEnableThisCode = true;
            rev = 0;
            x = 0;
            y = 0;
            lastShiftWasLimited = false;
            gear = -1;
            distanceIndicatorProgress = 0;
            score = 0;
            carSpeed = 0;
            resetHelper = true;
            for (int i = 0; i < 7; i++) {
                measuredReactionTime[i][0] = 0;
            }
            xPosBackground = 0;
            yPosBackground = 0;
            xPosCar = 10;
            yPosCar = 105;

            helpCalculateDrawBackground = millis();
            helpCalculateY = millis();
            helpCalculateReactionTime = millis();
            helpCalculateStartOfGameplayTime = millis();

            startUpGame = true;
            timePassed = millis();

            // Reset cached highscores
            for (int i = 0; i < 5; i++) {
                highscores[i] = 0;
            }

            pixels.clear();
            pixels.setPixelColor(0, pixels.Color(243,0,0));
            pixels.show();

            programState = 1;
            Serial.println("Info: Change to program state 'gameplay' completed.");
            break;
        case 2:
            Serial.println("Info: Preparing for program state change to 'scoreboard'.");
            newHighscorePosition = -1;
            loadHighscores(selectedCarIndex);
            addHighscore();
            programState = 2;
            Serial.println("Info: Change to program state 'scoreboard' completed.");
            break;
        default:
            Serial.println("Warning: Trying to change program state to an out of range index.");
    }
}

/*!
 * @brief Loads the user settings config as JSON from the filesystem and sets our global variables.
 * @return True, if the file was found and was parsable.
 */
bool loadUserSettings() {

    Serial.println("Start loading of user settings.");

    // Open file for reading
    File file = arcada.open(USERSETTINGSPATH);
    if (!file) {
        Serial.println(F("Failed to open user settings file"));
        return false;
    }

    // Deserialize the JSON document
    DeserializationError error = deserializeJson(configJSON, file);
    if (error) {
        Serial.println(F("Failed to read user settings file"));
        return false;
    }

    // Close the file (File's destructor doesn't close the file)
    file.close();

    brightnessDisplayIndex = configJSON["brightnessDisplay"] > 2 ? 0 : configJSON["brightnessDisplay"];
    brightnessLEDIndex = configJSON["brightnessLED"] > 3 ? 2 : configJSON["brightnessLED"];
    volumeOn = configJSON["volume"];

    Serial.println("Completed loading of user settings.");
    return true;
}

/**
 * @brief Saves the user settings config as JSON to the filesystem.
 * @return True, if the file was writable and saved.
 */
bool saveUserSettings() {

    Serial.println("Start saving of user settings.");

    configJSON["brightnessDisplay"] = brightnessDisplayIndex;
    configJSON["brightnessLED"] = brightnessLEDIndex;
    configJSON["volume"] = volumeOn;

    // Delete existing file, otherwise the configuration is appended to the file
    arcada.remove(USERSETTINGSPATH);

    // Open file for writing
    File file = arcada.open(USERSETTINGSPATH, O_WRITE | O_CREAT);
    if (!file) {
        Serial.println(F("Error: Failed to create user settings file"));
        return false;
    }

    // Serialize JSON to file
    if (serializeJson(configJSON, file) == 0) {
        Serial.println(F("Error: Failed to write to user settings file"));
        return false;
    }

    // Close the file (File's destructor doesn't close the file)
    file.close();

    Serial.println("Completed saving of user settings.");
    return true;
}

/**
 * @brief Draws some text on bottom which indicates possible button-actions
 */
void drawButtonHint(uint16_t color, String textFirst, String textSecond) {
    GFXcanvas16 *canvas = arcada.getCanvas();
    canvas->setTextColor(color);
    canvas->setTextSize(1);
    canvas->setCursor(4, 108);
    canvas->print(textFirst);
    canvas->setCursor(4, 118);
    canvas->print(textSecond);
}

/**
 * @brief Sets the correct amount of Neopixels to green for use in the main menu
 * @param amount Number of how many pixels show be activated
 */
void setMenuLEDs(uint8_t amount) {
    pixels.clear();
    for (uint8_t i = 0; i < amount; i++) {
        pixels.setPixelColor(i, pixels.Color(0, 255, 0));
    }
}


//////////////////////////////////////////////// Handle menu ////////////////////////////////////////////////

/**
 * @brief Draws a certain level of the main menu onto the canvas.
 * @param array An array of Strings that store the menu rows to be displayed.
 * @param size The size of the prior array.
 * @param index The index of the current selected element. Should be >0 and <size-1. 
 */
void drawMenu(String array[], int size, int index) {

    // inital setup
    GFXcanvas16 *canvas = arcada.getCanvas();
    canvas->fillScreen(ARCADA_BLACK);
    canvas->setTextColor(ARCADA_WHITE);
    canvas->setTextSize(1);

    // draw upper menu items
    if (index - 3 >= 0) {
        canvas->drawChar(40, 30, 0x1E, ARCADA_WHITE, ARCADA_BLACK, 1);
    }
    if (index - 2 >= 0) {
        canvas->setCursor(12, 40);
        canvas->print(array[index - 2]);
    }
    if (index - 1 >= 0) {
        canvas->setCursor(12, 50);
        canvas->print(array[index - 1]);
    }

    // left arrow if back action is available
    if (menuIndexSecondLevel != -1) {
        canvas->drawChar(2, 60, 0x11, ARCADA_WHITE, ARCADA_BLACK, 1);
    }

    // draw center/selected menu item
    canvas->setCursor(12, 60);
    canvas->print(array[index]);
    canvas->drawChar((array[index].length() * 6) + 17, 60, 0x10, ARCADA_WHITE, ARCADA_BLACK, 1);

    // draw lower menu items
    if (index + 1 <= size - 1) {
        canvas->setCursor(12, 70);
        canvas->print(array[index + 1]);
    }
    if (index + 2 <= size - 1) {
        canvas->setCursor(12, 80);
        canvas->print(array[index + 2]);
    }
    if (index + 3 <= size - 1) {
        canvas->drawChar(40, 90, 0x1F, ARCADA_WHITE, ARCADA_BLACK, 1);
    }

    // Draw game logo in the upper right corner
    canvas->setTextSize(2);
    canvas->setCursor(102, 4);
    canvas->print("RACE");
    canvas->setTextSize(1);
    canvas->setCursor(150, 4);
    canvas->print("2");
}

/**
 * @brief Handles the main menu part of the game on the top most layer.
 */
void handleMenu() {

    // Set neopixels
    setMenuLEDs(1);

    // Call submenus
    if (menuIndexSecondLevel == -1) {

        // Handle button presses
        if (buttonPressed(ARCADA_BUTTONMASK_UP) && menuIndexFirstLevel > 0) {
            menuIndexFirstLevel -= 1;
        } else if (buttonPressed(ARCADA_BUTTONMASK_DOWN) && menuIndexFirstLevel < MAINMENUSIZE - 1) {
            menuIndexFirstLevel += 1;
        } else if (buttonPressed(ARCADA_BUTTONMASK_A) || buttonPressed(ARCADA_BUTTONMASK_RIGHT)) {
            switch (menuIndexFirstLevel) {
                case 0:
                case 1:
                case 2:
                    menuIndexSecondLevel = 0;
                    break;
                case 3:
                    menuIndexSecondLevel = 0;
                    loadHighscores(selectedScoreboardCarIndex);
                    break;
                case 4:
                    menuIndexSecondLevel = 0;
                    creditsIndex = 0;
                    break;
                default:
                    Serial.println("Warning: Main menu index out of range.");
                    break;
            }
        }

        drawMenu(mainMenu, MAINMENUSIZE, menuIndexFirstLevel);
    } else {
        // Call submenus
        switch (menuIndexFirstLevel) {
            case 0:
                handleMenuPlay();
                break;
            case 1:
                handleMenuGarage();
                break;
            case 2:
                handleMenuSettings();
                break;
            case 3:
                handleScoreboardMenu();
                break;
            case 4:
                handleCredits();
                break;
            default:
                Serial.println("Warning: Main menu index out of range.");
                break;
        }
    }

    // Show pixels set in called submenus
    pixels.show();
}

/**
 * @brief Handles the main menu in category 'Play'.
 */
void handleMenuPlay() {

    // Set neopixels
    setMenuLEDs(2);
    
    if (menuIndexThirdLevel == -1) {

        drawMenu(playMenu, PLAYMENUSIZE, menuIndexSecondLevel);

        // Handle button presses
        if (buttonPressed(ARCADA_BUTTONMASK_UP) && menuIndexSecondLevel > 0) {
            menuIndexSecondLevel -= 1;
        } else if (buttonPressed(ARCADA_BUTTONMASK_DOWN) && menuIndexSecondLevel < SETTINGSMENUSIZE - 1) {
            menuIndexSecondLevel += 1;
        } else if (buttonPressed(ARCADA_BUTTONMASK_A) || buttonPressed(ARCADA_BUTTONMASK_RIGHT)) {
            switch (menuIndexSecondLevel) {
                case 0:
                    menuIndexThirdLevel = 0;
                    selectedBackground = 0;
                    break;
                case 1:
                    menuIndexThirdLevel = 0;
                    selectedBackground = 1;
                    break;
                case 2:
                    menuIndexThirdLevel = 0;
                    selectedBackground = 2;
                    break;
                default:
                    Serial.println("Warning: Play menu index out of range.");
                    break;
            }
        } else if (buttonPressed(ARCADA_BUTTONMASK_B) || buttonPressed(ARCADA_BUTTONMASK_LEFT)) {
            menuIndexSecondLevel = -1;
        }
    } else {
        changeProgramState(1);
    }
}

/**
 * @brief Handles the main menu in category 'Garage'.
 */
void handleMenuGarage() {

    // Set neopixels
    setMenuLEDs(2);

    // Draw background and car
    GFXcanvas16 *canvas = arcada.getCanvas();
    canvas->drawRGBBitmap(0, 0, background_garage, 160, 128);
    canvas->drawRGBBitmap(45, 90, carGraphics[selectedCarIndex], carMasks[selectedCarIndex], 72, 21);

    // Draw text
    canvas->setTextColor(ARCADA_WHITE);
    canvas->setTextSize(1);
    canvas->setCursor(42, 10);
    canvas->print("Selected car:");
    canvas->setCursor(80 - ((4 + carNames[selectedCarIndex].length()) * 3), 26);
    canvas->print("< " + carNames[selectedCarIndex] + " >");

    // Draw button information
    drawButtonHint(ARCADA_WHITE, "", "B - Back");

    // Handle button presses
    if (buttonPressed(ARCADA_BUTTONMASK_LEFT)) {
        if (selectedCarIndex <= 0) {
            selectedCarIndex = 2;
        } else {
            selectedCarIndex -= 1;
        }
    } else if (buttonPressed(ARCADA_BUTTONMASK_RIGHT)) {
        if (selectedCarIndex >= 2) {
            selectedCarIndex = 0;
        } else {
            selectedCarIndex += 1;
        }
    } else if (buttonPressed(ARCADA_BUTTONMASK_B)) {
        menuIndexSecondLevel = -1;
    }
}

/**
 * @brief Handles the main menu in category 'Settings'.
 */
void handleMenuSettings() {

    // Set neopixels
    setMenuLEDs(2);
    
    if (menuIndexThirdLevel == -1) {

        drawMenu(settingsMenu, SETTINGSMENUSIZE, menuIndexSecondLevel);

        // Handle button presses
        if (buttonPressed(ARCADA_BUTTONMASK_UP) && menuIndexSecondLevel > 0) {
            menuIndexSecondLevel -= 1;
        } else if (buttonPressed(ARCADA_BUTTONMASK_DOWN) && menuIndexSecondLevel < SETTINGSMENUSIZE - 1) {
            menuIndexSecondLevel += 1;
        } else if (buttonPressed(ARCADA_BUTTONMASK_A) || buttonPressed(ARCADA_BUTTONMASK_RIGHT)) {
            switch (menuIndexSecondLevel) {
                case 0:
                case 1:
                case 2:
                case 3:
                    menuIndexThirdLevel = 0;
                    break;
                default:
                    Serial.println("Warning: Brightness menu index out of range.");
                    break;
            }
        } else if (buttonPressed(ARCADA_BUTTONMASK_B) || buttonPressed(ARCADA_BUTTONMASK_LEFT)) {
            menuIndexSecondLevel = -1;
        }
    } else {
        // Call submenus
        switch (menuIndexSecondLevel) {
            case 0:
                handleBrightnessDisplay();
                break;
            case 1:
                handleBrightnessLed();
                break;
            case 2:
                handleVolume();
                break;
            case 3:
                resetScoreboard();
                break;
            default:
                Serial.println("Warning: Brightness menu index out of range.");
                break;
        }
    }
}

/**
 * @brief Handles the viewing of the scoreboard in the main menu
 */
void handleScoreboardMenu() {

    // Set neopixels
    setMenuLEDs(2);

    drawScoreboard(false, selectedScoreboardCarIndex);

    // Draw button information
    drawButtonHint(ARCADA_WHITE, "", "B - Back");

    // Handle button presses
    if (buttonPressed(ARCADA_BUTTONMASK_LEFT)) {
        if (selectedScoreboardCarIndex <= 0) {
            selectedScoreboardCarIndex = 2;
        } else {
            selectedScoreboardCarIndex -= 1;
        }
        loadHighscores(selectedScoreboardCarIndex);
    } else if (buttonPressed(ARCADA_BUTTONMASK_RIGHT)) {
        if (selectedScoreboardCarIndex >= 2) {
            selectedScoreboardCarIndex = 0;
        } else {
            selectedScoreboardCarIndex += 1;
        }
        loadHighscores(selectedScoreboardCarIndex);
    } else if (buttonPressed(ARCADA_BUTTONMASK_B)) {
        menuIndexSecondLevel = -1;
    }
}

/**
 * @brief Handles the display of the credits in the main menu
 */
void handleCredits() {

    // Set neopixels
    setMenuLEDs(2);
    
    // Draw text and scroll indicators
    GFXcanvas16 *canvas = arcada.getCanvas();
    canvas->fillScreen(ARCADA_BLACK);
    canvas->setTextColor(ARCADA_WHITE);
    canvas->setTextSize(1);
    if (creditsIndex != 0) {
        canvas->drawChar(77, 4, 0x1E, ARCADA_WHITE, ARCADA_BLACK, 1);
    }
    if (creditsIndex != 10) {
        canvas->drawChar(77, 106, 0x1F, ARCADA_WHITE, ARCADA_BLACK, 1);
    }
    for (int i = 0; i < 7; i++) {
        canvas->setCursor(4, 16 + (12 * i));
        canvas->print(creditsTexts[i + creditsIndex]);
    }

    // Draw button information
    drawButtonHint(ARCADA_WHITE, "", "B - Back");

    // Handle button presses
    if (buttonPressed(ARCADA_BUTTONMASK_UP) && creditsIndex > 0) {
        creditsIndex -= 1;
    } else if (buttonPressed(ARCADA_BUTTONMASK_DOWN) && creditsIndex < 10) {
        creditsIndex += 1;
    } else if (buttonPressed(ARCADA_BUTTONMASK_B)) {
        menuIndexSecondLevel = -1;
    }
}

/**
 * @brief Handles the brightness of the display, ranging from 10 to 100%
 */
void handleBrightnessDisplay() {

    // Set neopixels
    setMenuLEDs(3);
    
    GFXcanvas16 *canvas = arcada.getCanvas();
    canvas->fillScreen(ARCADA_BLACK);

    arcada.setBacklight(brightnessDisplayValues[brightnessDisplayIndex]);

    // print current value of brightness on screen
    canvas->setTextColor(ARCADA_WHITE);
    uint8_t pos = 20;
    for (int i = 0; i < 3; i++) {
        if (brightnessDisplayIndex == i) {
            canvas->setTextSize(2);
            canvas->setCursor(4, pos);
            canvas->print(brightnessDisplayLabels[i]);
            pos += 24;
        } else {
            canvas->setTextSize(1);
            canvas->setCursor(4, pos);
            canvas->print(brightnessDisplayLabels[i]);
            pos += 16;
        }
    }

    // Draw button information
    drawButtonHint(ARCADA_WHITE, "", "B - Save");

    // Handle button presses
    if (buttonPressed(ARCADA_BUTTONMASK_UP)) {
        if (brightnessDisplayIndex > 0) {
            brightnessDisplayIndex -= 1;
        }
    } else if (buttonPressed(ARCADA_BUTTONMASK_DOWN)) {
        if (brightnessDisplayIndex < 2) {
            brightnessDisplayIndex += 1;
        }
    } else if (buttonPressed(ARCADA_BUTTONMASK_B)) {
        menuIndexThirdLevel = -1;
        saveUserSettings();
    }
}

/**
 * @brief Handles the brightness of the LED's, ranging from 10 to 100%
 */
void handleBrightnessLed() {

    // Set the LED's brightness
    pixels.clear();
    pixels.setPixelColor(0, pixels.Color(255, 0, 0));
    pixels.setPixelColor(1, pixels.Color(255, 153, 0));
    pixels.setPixelColor(2, pixels.Color(255, 255, 0));
    pixels.setPixelColor(3, pixels.Color(0, 255, 0));
    pixels.setPixelColor(4, pixels.Color(0, 0, 255));
    pixels.setBrightness(brightnessLEDValues[brightnessLEDIndex]);
    
    GFXcanvas16 *canvas = arcada.getCanvas();
    canvas->fillScreen(ARCADA_BLACK);

    // print current value of brightness on screen
    canvas->setTextColor(ARCADA_WHITE);
    uint8_t pos = 20;
    for (int i = 0; i < 4; i++) {
        if (brightnessLEDIndex == i) {
            canvas->setTextSize(2);
            canvas->setCursor(4, pos);
            canvas->print(brightnessLEDLabels[i]);
            pos += 24;
        } else {
            canvas->setTextSize(1);
            canvas->setCursor(4, pos);
            canvas->print(brightnessLEDLabels[i]);
            pos += 16;
        }
    }

    // Draw button information
    drawButtonHint(ARCADA_WHITE, "", "B - Save");

    // Handle button presses
    if (buttonPressed(ARCADA_BUTTONMASK_UP)) {
        if (brightnessLEDIndex > 0) {
            brightnessLEDIndex -= 1;
        }
    } else if (buttonPressed(ARCADA_BUTTONMASK_DOWN)) {
        if (brightnessLEDIndex < 3) {
            brightnessLEDIndex += 1;
        }
    } else if (buttonPressed(ARCADA_BUTTONMASK_B)) {
        menuIndexThirdLevel = -1;
        saveUserSettings();
    }
}

/**
 * @brief Handles the volume of the speaker (turns it on or off)
 */
void handleVolume() {

    // Set neopixels
    setMenuLEDs(3);

    GFXcanvas16 *canvas = arcada.getCanvas();
    canvas->fillScreen(ARCADA_BLACK);

    // print current state of volume on screen
    canvas->setTextColor(ARCADA_WHITE);
    canvas->setCursor(4, 20);
    if (volumeOn) {
        canvas->setTextSize(2);
        canvas->print(volumeLabels[0]);
        canvas->setTextSize(1);
        canvas->setCursor(4, 44);
    } else {
        canvas->setTextSize(1);
        canvas->print(volumeLabels[0]);
        canvas->setTextSize(2);
        canvas->setCursor(4, 36);
    }
    canvas->print(volumeLabels[1]);

    // Draw button information
    drawButtonHint(ARCADA_WHITE, "", "B - Save");

    // Handle button presses
    if (buttonPressed(ARCADA_BUTTONMASK_UP)) {
        volumeOn = true;
    } else if (buttonPressed(ARCADA_BUTTONMASK_DOWN)) {
        volumeOn = false;
    } else if (buttonPressed(ARCADA_BUTTONMASK_B)) {
        menuIndexThirdLevel = -1;
        saveUserSettings();
    }
    
    arcada.enableSpeaker(volumeOn);
}

/**
 * @brief Handles the volume of the speaker (turns it on or off)
 */
void resetScoreboard() {

    // Set neopixels
    setMenuLEDs(3);

    GFXcanvas16 *canvas = arcada.getCanvas();
    canvas->fillScreen(ARCADA_BLACK);
    canvas->setTextColor(ARCADA_WHITE);
    canvas->setTextSize(1);
    canvas->setCursor(4, 4);
    canvas->print("Reset Scoreboard");
    canvas->setCursor(4, 16);
    canvas->print("for all vehicles?");
    canvas->setCursor(4, 30);
    canvas->print("(This process may take");
    canvas->setCursor(4, 42);
    canvas->print("a second)");

    // Draw button information
    drawButtonHint(ARCADA_WHITE, "A - Confirm", "B - Cancel");

    // Handle button presses
    if (buttonPressed(ARCADA_BUTTONMASK_A)) {
        for (int i = 0; i < 5; i++) {
            highscores[i] = 0;
        }
        for (int i = 0; i < 3; i++) {
            saveHighscores(i);
        }
        menuIndexThirdLevel = -1;
    } else if (buttonPressed(ARCADA_BUTTONMASK_B)) {
        menuIndexThirdLevel = -1;
    }
}


//////////////////////////////////////////////// Handle gameplay ////////////////////////////////////////////////


/**
 * @brief Wraps around the actual gameplay and adds a countdown on the beginning as well as a smooth ending transition
 */
void handleGameplayWrapper() {
    GFXcanvas16 *canvas = arcada.getCanvas();

    // Countdown to begin the gameplay
    if (startUpGame == true) {
        drawBackground(false);
        drawCar();
        drawIngameStats();
        canvas->setTextSize(2);
        canvas->setTextColor(ARCADA_YELLOW);
        canvas->setCursor(75, 56);

        // Draw the countdown numbers
        if (millis() > timePassed + 3000) {
            startUpGame = false;
            gear = 0;
            helpCalculateDistanceDriven = millis();
        } else if (millis() > timePassed + 2000) {
            canvas->print("1");
        } else if (millis() > timePassed + 1000) {
            canvas->print("2");
        } else if (millis() > timePassed) {
            canvas->print("3");
        }
    } else {

        // Actual gameplay
        if (distanceDriven < maxDistance) {
            if(resetHelper){
                helpCalculateStartOfGameplayTime = millis();
                resetHelper = false;
            }
            handleGameplay();
        } else {

            // Make car move out of screen and handle switch to scoreboard
            if (buttonPressed(ARCADA_BUTTONMASK_A)) {
                changeProgramState(2);
            }

            drawBackground(false);
            drawCar();
            drawIngameStats();
            drawButtonHint(ARCADA_YELLOW, "", "A - Continue");
        }
    }
}

/**
 * @brief Handles the gameplay part of the game.
 */
void handleGameplay() {

    if (distanceDriven < maxDistance) {
        if (lastShiftWasLimited) {
            y = y + millis() - helpCalculateY;
            lastShiftWasLimited = false;
        } else {
            y = y;
        }
        x = ((millis() - helpCalculateStartOfGameplayTime - y) / 1000)*multiplyxwith[selectedCarIndex];

        // 1 LED
        if (carSpeed < 0.65 * maxSpeedPerGear[selectedCarIndex][gear]) {
            carSpeed = f1*x*x*x - f2*x*x + f3*x + f4;
            distanceDriven = ((distanceDriven + (carSpeed / 3.6) * ((millis() - helpCalculateDistanceDriven)) / 1000));
            helpCalculateDistanceDriven = millis();
            if (carSpeed >= maxSpeedPerGear[selectedCarIndex][gear]) {
                carSpeed = maxSpeedPerGear[selectedCarIndex][gear];
                helpCalculateY = millis();
                lastShiftWasLimited = true;
            }

            rev = 1;
        }

        // 2 LEDs
        if (carSpeed >= 0.65 * maxSpeedPerGear[selectedCarIndex][gear] && carSpeed < 0.72 * maxSpeedPerGear[selectedCarIndex][gear]) {
            carSpeed = f1*x*x*x - f2*x*x + f3*x + f4;
            distanceDriven = ((distanceDriven + (carSpeed / 3.6) * ((millis() - helpCalculateDistanceDriven)) / 1000));
            helpCalculateDistanceDriven = millis();
            if (carSpeed >= maxSpeedPerGear[selectedCarIndex][gear]) {
                carSpeed = maxSpeedPerGear[selectedCarIndex][gear];
                helpCalculateY = millis();
                lastShiftWasLimited = true;
            }

            rev = 2;
        }

        // 3 LEDs
        if (carSpeed >= 0.72 * maxSpeedPerGear[selectedCarIndex][gear] && carSpeed < 0.80 * maxSpeedPerGear[selectedCarIndex][gear]) {
            if (buttonPressed(ARCADA_BUTTONMASK_A) && gear < carGearsAmount[selectedCarIndex] - 1) {
                measuredReactionTime[gear][0] = 9999; 
                carSpeed = (f1*x*x*x - f2*x*x + f3*x + f4)/(1 + 0.0050 * (0.8 * maxSpeedPerGear[selectedCarIndex][gear] - carSpeed));//Calculates speed difference to the green light in km/h. Function f is divided by a, thus stretching the function for one calculation
                score = score + 5000 - (tooEarlyVariable[selectedCarIndex][gear] * (0.8 * maxSpeedPerGear[selectedCarIndex][gear] - carSpeed)); //Calculates score for shifting too early(yellow light). For each shift too early, you get 5000 points and lose 5000 points divided by the speed in km/h that one can shift too early divided by the speed that was shifted too early: 5000-(5000/(kmhAbleTooEarly/kmhActuallyTooEarly))

                gear = gear + 1;
                temporarilyEnableThisCode = false;
                helpCalculateReactionTime = millis(); 
            }

            if (temporarilyEnableThisCode) {
                carSpeed = f1*x*x*x - f2*x*x + f3*x +f4;
            }

            distanceDriven = ((distanceDriven + (carSpeed / 3.6) * ((millis() - helpCalculateDistanceDriven)) / 1000));
            helpCalculateDistanceDriven = millis();
            if (carSpeed >=maxSpeedPerGear[selectedCarIndex][gear]) {
                carSpeed = maxSpeedPerGear[selectedCarIndex][gear];
                helpCalculateY = millis();
                lastShiftWasLimited = true;
            }
            rev = 3;

            temporarilyEnableThisCode = true;
            helpCalculateReactionTime = millis();
        }

        // 4 LEDs
        if (carSpeed >= 0.80 * maxSpeedPerGear[selectedCarIndex][gear] && carSpeed < 0.88 * maxSpeedPerGear[selectedCarIndex][gear]) {

            carSpeed = f1*x*x*x - f2*x*x + f3*x + f4;

            if (buttonPressed(ARCADA_BUTTONMASK_A) && gear < carGearsAmount[selectedCarIndex] - 1) {

            //if (carSpeed >= 0.82 * maxSpeedPerGear[selectedCarIndex][gear] && gear < 7 && maxSpeedPerGear[selectedCarIndex][gear+1] > -0.5) {
                measuredReactionTime[gear][0] = millis() - helpCalculateReactionTime; 
                score = score + 5000 - 1 * (measuredReactionTime[gear][0]); //Calculates score for shifting on time or later(green light). For each shift you get 5000 points, for each millisecond too late frome the perfect moment, you lose one point-
                gear = gear + 1;
                helpCalculateReactionTime = millis(); 
            }

            distanceDriven = ((distanceDriven + (carSpeed / 3.6) * ((millis() - helpCalculateDistanceDriven)) / 1000));
            helpCalculateDistanceDriven = millis();
            if (carSpeed >= maxSpeedPerGear[selectedCarIndex][gear]) {
                carSpeed = maxSpeedPerGear[selectedCarIndex][gear];
                helpCalculateY = millis();
                lastShiftWasLimited = true;
            }
            rev = 4;
        }

        // 5 LEDs
        if (carSpeed >= 0.88 * maxSpeedPerGear[selectedCarIndex][gear]) {
            if (buttonPressed(ARCADA_BUTTONMASK_A) && gear < carGearsAmount[selectedCarIndex] - 1) {
                measuredReactionTime[gear][0] = millis() - helpCalculateReactionTime;
                carSpeed = (f1*x*x*x - f2*x*x + f3*x + f4) / (1 + 0.0005 * (carSpeed - 0.8 * maxSpeedPerGear[selectedCarIndex][gear]));
                distanceDriven = ((distanceDriven + (carSpeed / 3.6) * ((millis() - helpCalculateDistanceDriven)) / 1000));
                helpCalculateDistanceDriven = millis();
                if (carSpeed >= maxSpeedPerGear[selectedCarIndex][gear]) {
                    carSpeed = maxSpeedPerGear[selectedCarIndex][gear];
                    helpCalculateY = millis();
                    lastShiftWasLimited = true;
                }
                score = score + 5000 - 1 * (measuredReactionTime[gear][0]); //Calculates score for shifting after the 3rd red light. For each shift you get 5000 points, for each millisecond too late frome the perfect moment, you lose one point-
                gear = gear + 1;
                temporarilyEnableThisCode = false;
                helpCalculateReactionTime = millis(); 

                if (carSpeed > maxSpeedPerGear[selectedCarIndex][gear]) {
                    carSpeed = maxSpeedPerGear[selectedCarIndex][gear];
                    helpCalculateY = millis();
                    lastShiftWasLimited = true;
                }

                rev = 5;
            }

            if (temporarilyEnableThisCode) {
                carSpeed = f1*x*x*x - f2*x*x + f3*x + f4;
                distanceDriven = ((distanceDriven + (carSpeed / 3.6) * ((millis() - helpCalculateDistanceDriven)) / 1000));
                helpCalculateDistanceDriven = millis();
                if (carSpeed > maxSpeedPerGear[selectedCarIndex][gear]) {
                    carSpeed = maxSpeedPerGear[selectedCarIndex][gear];
                    helpCalculateY = millis();
                    lastShiftWasLimited = true;
                }

                rev = 5;
            }
            temporarilyEnableThisCode = true;
        }
        drawBackground(true);
        drawCar();
        drawIngameStats();
        revCounter();
    }
}

/**
 * @brief Turns on the Neopixels according to the current rev number
 */
void revCounter() {
    pixels.clear();
    switch (rev) {
        case 5:
            pixels.setPixelColor(4, pixels.Color(243,0,0)); // red
        case 4:
            pixels.setPixelColor(3, pixels.Color(72,251,0)); // green
        case 3:
            pixels.setPixelColor(2, pixels.Color(249,233,9)); // yellow
        case 2:
            pixels.setPixelColor(1, pixels.Color(243,0,0)); // red
        case 1:
            pixels.setPixelColor(0, pixels.Color(243,0,0)); // red
            break;
        default:
            Serial.println("Warning: revCounter() called with out of bonds number.");
            break;
    }
    pixels.show();
}

/**
 * @brief Calculates the average Reaction time
 * @return the reaction time
 */

int calculateAverageReactionTime() {
    int8_t maxGearIndex = -1;
    switch (selectedCarIndex) {
        case 0:
            maxGearIndex = 6;
            break;
        case 1:
            maxGearIndex = 5;
            break;
        case 2:
            maxGearIndex = 7;
            break;
        default:
            Serial.print("Warning: calculateAverageReactionTime() called with selectedCarIndex out of bonds.");
            break;
    }
    for(int i = 0; i < maxGearIndex; i++){
        averageReactionTime = averageReactionTime + measuredReactionTime[i][0];
        averageReactionTime = averageReactionTime / maxGearIndex;
    }
    return averageReactionTime;
}

/**
 * @brief Draws ingame stats on screen, including car speed, gear and distance progression
 */
void drawIngameStats() {
    GFXcanvas16 *canvas = arcada.getCanvas();

    // Draw distance progression background
    canvas->setTextColor(ARCADA_BLACK);
    canvas->setTextSize(1);
    for (int i = 0; i < 99; i++) {
        canvas->setCursor(i + 30, 20);
        canvas->print(".");
    }

    // Draw distance progression current
    canvas->setTextColor(ARCADA_YELLOW);
    distanceIndicatorProgress = (distanceDriven / maxDistance) * 100;
    for (double i = 0; i < distanceIndicatorProgress; i++) {
        canvas->setCursor(i + 30, 20);
        canvas->print(".");
    }

    // Draw speed
    canvas->setTextSize(1);
    canvas->setCursor(4, 4);
    canvas->print(String((int)(carSpeed)) + " km/h");

    // Draw gear
    canvas->setCursor(130, 4);
    canvas->print((gear == -1 ? "N" : String(gear + 1)) + "/" + String(carGearsAmount[selectedCarIndex]));
}

/**
 * @brief Draws the background image and moves it to create an illusion of speed
 * @param moving True if background is supposed to move, false if background is supposed to be static
 */
void drawBackground(bool moving) {
    GFXcanvas16 *canvas = arcada.getCanvas();

    if (millis() - helpCalculateDrawBackground && moving) {
        xPosBackground = (int) (xPosBackground - (carSpeed * 0.035));

        if (xPosBackground < -160) {
            xPosBackground = xPosBackground + 160;
        }
        helpCalculateDrawBackground = millis();

        if (selectedBackground == 0) {
            canvas->drawRGBBitmap(xPosBackground, yPosBackground, background_city, 160, 128);
            canvas->drawRGBBitmap(xPosBackground + 160, yPosBackground, background_city, 160, 128);
            canvas->drawRGBBitmap(xPosBackground + 320, yPosBackground, background_city, 160, 128);
        } else if (selectedBackground == 1) {
            canvas->drawRGBBitmap(xPosBackground, yPosBackground, background_fields, 160, 128);
            canvas->drawRGBBitmap(xPosBackground + 160, yPosBackground, background_fields, 160, 128);
            canvas->drawRGBBitmap(xPosBackground + 320, yPosBackground, background_fields, 160, 128);
        } else if (selectedBackground == 2) {
            canvas->drawRGBBitmap(xPosBackground, yPosBackground, background_beach, 160, 128);
            canvas->drawRGBBitmap(xPosBackground + 160, yPosBackground, background_beach, 160, 128);
            canvas->drawRGBBitmap(xPosBackground + 320, yPosBackground, background_beach, 160, 128);
        }

    } else {
        if (selectedBackground == 0) {
            canvas->drawRGBBitmap(0, yPosBackground, background_city, 160, 128);
        } else if (selectedBackground == 1) {
            canvas->drawRGBBitmap(0, yPosBackground, background_fields, 160, 128);
        } else if (selectedBackground == 2) {
            canvas->drawRGBBitmap(0, yPosBackground, background_beach, 160, 128);
        }
    }
}

/**
 * @brief Draws the car and moves it slightly while accelerating
 */
void drawCar() {
    GFXcanvas16 *canvas = arcada.getCanvas();

    // During gameplay
    if (distanceDriven < maxDistance && xPosCar < 250) {
        canvas->drawRGBBitmap(xPosCar + carSpeed / 10, yPosCar, carGraphics[selectedCarIndex], carMasks[selectedCarIndex], 72, 21);
    } else {
        // Move out of screen after gameplay
        xPosCar += 8;
        canvas->drawRGBBitmap(xPosCar + carSpeed / 10, yPosCar, carGraphics[selectedCarIndex], carMasks[selectedCarIndex], 72, 21);
    }
}


//////////////////////////////////////////////// Handle scoreboard ////////////////////////////////////////////////

/**
 * @brief Handles the scoreboard part of the game.
 */
void handleScoreboard() {
    pixels.clear();
    drawScoreboard(true, selectedCarIndex);

    // Draw button information
    drawButtonHint(ARCADA_WHITE, "", "A - To main menu");

    if (buttonPressed(ARCADA_BUTTONMASK_A)) {
        saveHighscores(selectedCarIndex);
        changeProgramState(0);
    }
}

/**
 * @brief Draws the scoreboard on the screen
 */
void drawScoreboard(bool postGame, uint8_t carIndex) {
    uint16_t lastThreeDigits = 0;
    GFXcanvas16 *canvas = arcada.getCanvas();
    canvas->fillScreen(ARCADA_BLACK);
    canvas->setTextColor(ARCADA_WHITE);
    canvas->setTextSize(1);

    // Draw car type and score
    if (postGame) {
        canvas->setCursor(26, 16);
        lastThreeDigits = score % 1000;
        canvas->print("Score: " + String((score - lastThreeDigits) / 1000) + "." + (lastThreeDigits >= 100 ? String(lastThreeDigits) : (lastThreeDigits >= 10 ? ("0" + String(lastThreeDigits)) : ("00" + String(lastThreeDigits)))));
        canvas->setCursor(80 - ((5 + carNames[carIndex].length()) * 3), 4);
        canvas->print("Car: " + carNames[carIndex]);
    } else {
        canvas->setCursor(80 - ((9 + carNames[carIndex].length()) * 3), 4);
        canvas->print("< Car: " + carNames[carIndex] + " >");
    }

    // Draw highscore table
    canvas->setCursor(26, 36);
    canvas->print("Highscores:");
    for (int i = 0; i < 5; i++) {
        canvas->setCursor(26, 48 + i * 12);
        if (highscores[i] == 0) {
            canvas->print("-");
        } else {
            lastThreeDigits = highscores[i] % 1000;
            canvas->print(String((highscores[i] - lastThreeDigits) / 1000) + "." + (lastThreeDigits >= 100 ? String(lastThreeDigits) : (lastThreeDigits >= 10 ? ("0" + String(lastThreeDigits)) : ("00" + String(lastThreeDigits)))));
            if (newHighscorePosition == i) {
                canvas->setCursor(4, 48 + i * 12);
                canvas->print("NEW");
            }
        }
    }
}

/**
 * @brief Inserts the current score into the highscores array at the right position
 */
void addHighscore() {
    if (score > highscores[4]) {
        highscores[4] = score;
        newHighscorePosition = 4;
        for (int i = 3; i >= 0; i--) {
            if (highscores[i] < highscores[i + 1]) {
                uint16_t temp = highscores[i];
                highscores[i] = highscores[i + 1];
                highscores[i + 1] = temp;
                newHighscorePosition = i;
            }
        }
    }
}

/*!
 * @brief Loads the highscores as JSON from the filesystem and sets our global variables.
 * @return True, if the file was found and was parsable.
 */
bool loadHighscores(uint8_t carIndex) {

    Serial.println("Start loading of highscores.");

    // Open file for reading
    File file = arcada.open((HIGHSCORESPATH + String(carIndex) + JSONEXTENSION).c_str());
    if (!file) {
        Serial.println(F("Failed to open highscores file"));
        return false;
    }

    // Deserialize the JSON document
    DeserializationError error = deserializeJson(highscoreJSON, file);
    if (error) {
        Serial.println(F("Failed to read highscores file"));
        return false;
    }

    // Close the file (File's destructor doesn't close the file)
    file.close();

    highscores[0] = highscoreJSON["highscore1"];
    highscores[1] = highscoreJSON["highscore2"];
    highscores[2] = highscoreJSON["highscore3"];
    highscores[3] = highscoreJSON["highscore4"];
    highscores[4] = highscoreJSON["highscore5"];

    Serial.println("Completed loading of highscores.");
    return true;
}

/**
 * @brief Saves the highscores as JSON to the filesystem.
 * @return True, if the file was writable and saved.
 */
bool saveHighscores(uint8_t carIndex) {

    Serial.println("Start saving of highscores.");

    highscoreJSON["highscore1"] = highscores[0];
    highscoreJSON["highscore2"] = highscores[1];
    highscoreJSON["highscore3"] = highscores[2];
    highscoreJSON["highscore4"] = highscores[3];
    highscoreJSON["highscore5"] = highscores[4];

    // Delete existing file, otherwise the configuration is appended to the file
    arcada.remove((HIGHSCORESPATH + String(carIndex) + JSONEXTENSION).c_str());

    // Open file for writing
    File file = arcada.open((HIGHSCORESPATH + String(carIndex) + JSONEXTENSION).c_str(), O_WRITE | O_CREAT);
    if (!file) {
        Serial.println(F("Error: Failed to create highscores file"));
        return false;
    }

    // Serialize JSON to file
    if (serializeJson(highscoreJSON, file) == 0) {
        Serial.println(F("Error: Failed to write to highscores file"));
        return false;
    }

    // Close the file (File's destructor doesn't close the file)
    file.close();

    Serial.println("Completed saving of highscores.");
    return true;
}

//////////////////////////////////////////////// Audio ////////////////////////////////////////////////

/**
 * @brief Loads the music file from the flash storage
 */
void loadMusic() {

    // Open the first wav file in the music folder
    musicFile = arcada.openFileByIndex("/music", 0, O_READ, "wav");

    Serial.println("Loading sound file...");
  
    uint32_t sampleRate;

    // Load wav information, enable speakers and set interrupt for sampling
    wavStatus status = arcada.WavLoad(musicFile, &sampleRate);
   
    if ((status == WAV_LOAD) || (status == WAV_EOF)) {
        if (volumeOn == 1) {
            arcada.enableSpeaker(true);
            arcada.timerCallback(sampleRate, timedInterruptCallback);
        }
    } 
}

/**
 * @brief Uses interrupts to call a function as often as wished (pro second)
 */
void timedInterruptCallback(void) {

    if (!musicIsPlaying) {
        beatTime = millis() + bpmTiming;
    }

    // Output the next sample 
    wavStatus status = arcada.WavPlayNextSample();

    uint32_t timeNow = millis();

    if (timeNow >= beatTime + (bpmTiming >> 1)) {
        beatTime = beatTime + bpmTiming;
    }

    // Checks if the wav file has ended and stops music
    if (status == WAV_EOF) {
        arcada.timerStop();
        arcada.enableSpeaker(false);
        musicIsPlaying = false;
        musicFile.close();
    } else {
        musicIsPlaying = true;
    }
}
